import os

import requests

from grafap._auth import Decorators
from grafap._helpers import _basic_retry


@_basic_retry
@Decorators._refresh_graph_token
def get_sp_termstore_groups(site_id: str) -> dict:
    """
    Lists all termstore group objects in a site

    :param site_id: The site id
    """
    if "GRAPH_BASE_URL" not in os.environ:
        raise Exception("Error, could not find GRAPH_BASE_URL in env")

    try:
        response = requests.get(
            os.environ["GRAPH_BASE_URL"] + site_id + "/termStore/groups",
            headers={"Authorization": "Bearer " + os.environ["GRAPH_BEARER_TOKEN"]},
            timeout=30,
        )
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise Exception(
            f"Error {e.response.status_code}, could not get termstore groups: {e}"
        )
    except requests.exceptions.RequestException as e:
        raise Exception(f"Error, could not get termstore groups: {e}")

    return response.json()



import os

import requests

from grafap._auth import Decorators
from grafap._helpers import _basic_retry


@Decorators._refresh_graph_token
def get_sp_sites() -> dict:
    """
    Gets all site data in a given tenant
    """
    if "GRAPH_BASE_URL" not in os.environ:
        raise Exception("Error, could not find GRAPH_BASE_URL in env")

    @_basic_retry
    def recurs_get(url, headers):
        """
        Recursive function to handle pagination
        """
        try:
            response = requests.get(url, headers=headers, timeout=30)
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise Exception(
                f"Error {e.response.status_code}, could not get sharepoint site data: {e}"
            )
        except requests.exceptions.RequestException as e:
            raise Exception(f"Error, could not get sharepoint site data: {e}")

        data = response.json()

        # Check for the next page
        if "@odata.nextLink" in data:
            return data["value"] + recurs_get(data["@odata.nextLink"], headers)
        else:
            return data["value"]

    result = recurs_get(
        os.environ["GRAPH_BASE_URL"],
        headers={"Authorization": "Bearer " + os.environ["GRAPH_BEARER_TOKEN"]},
    )
    return result



import os
import sys
from datetime import datetime, timedelta
from pprint import pprint

from dotenv import load_dotenv

# Load env vars from a .env file
load_dotenv()

sys.path.insert(0, "")

from grafap import *

# SharePoint Sites

# sites = grafap.get_sp_sites()

# SharePoint Lists

# lists = grafap.get_sp_lists(sites[0]["id"])
# list_items = grafap.get_sp_list_items(sites[0]["id"], lists[0]["id"])
# list_item = grafap.get_sp_list_item(sites[0]["id"], lists[0]["id"], list_items[0]["id"])

# grafap.create_sp_item(
#     sites[0]["id"],
#     lists[0]["id"],
#     {
#         "Title": "Test",
#         "Description": "Test",
#     },
# )

# grafap.update_sp_item(
#     sites[0]["id"],
#     lists[0]["id"],
#     list_items[0]["id"],
#     {
#         "Title": "Test",
#         "Description": "Test",
#     },
# )

# grafap.delete_sp_item(sites[0]["id"], lists[0]["id"], list_items[0]["id"])

# SharePoint Site Users

# res = grafap.ensure_sp_user(
#     "SITE URL",
#     "email@domain.com",
# )

# List Attachments

# attachments = get_list_attachments(
#     os.environ["SP_SITE_INTERNAL"],
#     os.environ["SP_SITE_EXAMPLE_LIST_NAME"],
#     4,
#     download=True,
# )

# # Write first attachment data to a file
# with open(attachments[0]["name"], "wb") as f:
#     f.write(attachments[0]["data"])

# Document Library

# file = grafap.get_file(
#     "FULL FILE URL",
# )

pass

# AD Users

# users = grafap.get_ad_users(
#     select="id,userPrincipalName,givenName,surname,displayName,department,businessPhones,employeeOrgData,employeeId",
#     filter="mail eq 'example@domain.com'",
#     expand="manager($select=id,userPrincipalName,givenName,surname,displayName,department,businessPhones,employeeOrgData,employeeId,manager)",
# )



import base64
import hashlib
import os
import time
import uuid
from datetime import datetime, timedelta

import jwt
import requests
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import load_pem_private_key, pkcs12
from OpenSSL import crypto


class Decorators:
    """
    Decorators class for handling token refreshing
    for Microsoft Graph and Sharepoint Rest API
    """

    @staticmethod
    def _refresh_graph_token(decorated):
        """
        Decorator to refresh the graph access token if it has expired
        """

        def wrapper(*args, **kwargs):
            """
            Wrapper function
            """
            if "GRAPH_BEARER_TOKEN_EXPIRES_AT" not in os.environ:
                expires_at = "01/01/1901 00:00:00"
            else:
                expires_at = os.environ["GRAPH_BEARER_TOKEN_EXPIRES_AT"]
            if (
                "GRAPH_BEARER_TOKEN" not in os.environ
                or datetime.strptime(expires_at, "%m/%d/%Y %H:%M:%S") < datetime.now()
            ):
                Decorators._get_graph_token()
            return decorated(*args, **kwargs)

        wrapper.__name__ = decorated.__name__
        return wrapper

    @staticmethod
    def _refresh_sp_token(decorated):
        """
        Decorator to refresh the sharepoint rest API access token if it has expired
        """

        def wrapper(*args, **kwargs):
            """
            Wrapper function
            """
            if "SP_BEARER_TOKEN_EXPIRES_AT" not in os.environ:
                expires_at = "01/01/1901 00:00:00"
            else:
                expires_at = os.environ["SP_BEARER_TOKEN_EXPIRES_AT"]
            if (
                "SP_BEARER_TOKEN" not in os.environ
                or datetime.strptime(expires_at, "%m/%d/%Y %H:%M:%S") < datetime.now()
            ):
                Decorators._get_sp_token()
            return decorated(*args, **kwargs)

        wrapper.__name__ = decorated.__name__
        return wrapper

    @staticmethod
    def _get_graph_token():
        """
        Get Microsoft Graph bearer token
        """
        if "GRAPH_LOGIN_BASE_URL" not in os.environ:
            raise Exception("Error, could not find GRAPH_LOGIN_BASE_URL in env")
        if "GRAPH_TENANT_ID" not in os.environ:
            raise Exception("Error, could not find GRAPH_TENANT_ID in env")
        if "GRAPH_CLIENT_ID" not in os.environ:
            raise Exception("Error, could not find GRAPH_CLIENT_ID in env")
        if "GRAPH_CLIENT_SECRET" not in os.environ:
            raise Exception("Error, could not find GRAPH_CLIENT_SECRET in env")
        if "GRAPH_GRANT_TYPE" not in os.environ:
            raise Exception("Error, could not find GRAPH_GRANT_TYPE in env")
        if "GRAPH_SCOPES" not in os.environ:
            raise Exception("Error, could not find GRAPH_SCOPES in env")
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/x-www-form-urlencoded",
        }
        response = requests.post(
            os.environ["GRAPH_LOGIN_BASE_URL"]
            + os.environ["GRAPH_TENANT_ID"]
            + "/oauth2/v2.0/token",
            headers=headers,
            data={
                "client_id": os.environ["GRAPH_CLIENT_ID"],
                "client_secret": os.environ["GRAPH_CLIENT_SECRET"],
                "grant_type": os.environ["GRAPH_GRANT_TYPE"],
                "scope": os.environ["GRAPH_SCOPES"],
            },
            timeout=30,
        )
        try:
            os.environ["GRAPH_BEARER_TOKEN"] = response.json()["access_token"]
        except Exception as e:
            raise Exception(
                f"Error, could not set OS env bearer token: {response.content}"
            )
        try:
            expires_at = datetime.now() + timedelta(
                seconds=response.json()["expires_in"]
            )
            os.environ["GRAPH_BEARER_TOKEN_EXPIRES_AT"] = expires_at.strftime(
                "%m/%d/%Y %H:%M:%S"
            )
        except Exception as e:
            raise Exception(f"Error, could not set os env expires at: {e}")

    @staticmethod
    def _get_sp_token():
        """
        Gets Sharepoint Rest API bearer token.
        """
        if "SP_LOGIN_BASE_URL" not in os.environ:
            raise Exception("Error, could not find SP_LOGIN_BASE_URL in env")
        if "SP_TENANT_ID" not in os.environ:
            raise Exception("Error, could not find SP_TENANT_ID in env")
        if "SP_CLIENT_ID" not in os.environ:
            raise Exception("Error, could not find SP_CLIENT_ID in env")
        # if "SP_CLIENT_SECRET" not in os.environ:
        #     raise Exception("Error, could not find SP_CLIENT_SECRET in env")
        if "SP_CERTIFICATE_PATH" not in os.environ:
            raise Exception("Error, could not find SP_CERTIFICATE_PATH in env")
        if "SP_CERTIFICATE_PASSWORD" not in os.environ:
            raise Exception("Error, could not find SP_CERTIFICATE_PASSWORD in env")
        if "SP_GRANT_TYPE" not in os.environ:
            raise Exception("Error, could not find SP_GRANT_TYPE in env")
        if "SP_SITE" not in os.environ:
            raise Exception("Error, could not find SP_SITE in env")

        # Load the certificate
        with open(os.environ["SP_CERTIFICATE_PATH"], "rb") as cert_file:
            cert_data = cert_file.read()
        pfx = pkcs12.load_key_and_certificates(
            cert_data, str.encode(os.environ["SP_CERTIFICATE_PASSWORD"])
        )

        # Extract the private key and certificate
        private_key = pfx[0]
        certificate = pfx[1]

        private_key_pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )

        # Compute the SHA-1 thumbprint of the certificate
        cert_der = certificate.public_bytes(serialization.Encoding.DER)
        thumbprint = hashlib.sha1(cert_der).digest()
        thumbprint_b64 = (
            base64.urlsafe_b64encode(thumbprint).decode("utf-8").rstrip("=")
        )

        # JWT payload
        payload = {
            "aud": f"https://login.microsoftonline.com/{os.environ['GRAPH_TENANT_ID']}/oauth2/v2.0/token",
            "iss": os.environ["GRAPH_CLIENT_ID"],
            "sub": os.environ["GRAPH_CLIENT_ID"],
            "jti": str(uuid.uuid4()),
            "exp": int(time.time()) + 600,
        }

        # JWT header with x5t thumbprint
        headers = {"x5t": thumbprint_b64}

        # Generate the JWT assertion
        jwt_assertion = jwt.encode(
            payload, private_key_pem, algorithm="RS256", headers=headers
        )

        response = requests.post(
            os.environ["SP_LOGIN_BASE_URL"]
            + os.environ["SP_TENANT_ID"]
            + "/oauth2/v2.0/token",
            headers=headers,
            data={
                "client_id": os.environ["SP_CLIENT_ID"],
                "grant_type": os.environ["SP_GRANT_TYPE"],
                "scope": os.environ["SP_SCOPES"],
                "client_assertion_type": "urn:ietf:params:oauth:client-assertion-type:jwt-bearer",
                "client_assertion": jwt_assertion,
            },
            timeout=30,
        )

        try:
            os.environ["SP_BEARER_TOKEN"] = response.json()["access_token"]
        except Exception as e:
            raise Exception(f"Error, could not set OS env bearer token: {e}")
        try:
            expires_at = datetime.now() + timedelta(
                seconds=float(response.json()["expires_in"])
            )
            os.environ["SP_BEARER_TOKEN_EXPIRES_AT"] = expires_at.strftime(
                "%m/%d/%Y %H:%M:%S"
            )
        except Exception as e:
            raise Exception(f"Error, could not set os env expires at: {e}")



import os
from typing import Any, Dict, Literal
from urllib.parse import urlparse

import requests

from grafap._auth import Decorators
from grafap._helpers import _basic_retry, _fetch_page


@Decorators._refresh_graph_token
def get_sp_lists(site_id: str) -> dict:
    """
    Gets all lists in a given site

    :param site_id: The site id to get lists from
    """
    if "GRAPH_BASE_URL" not in os.environ:
        raise Exception("Error, could not find GRAPH_BASE_URL in env")

    @_basic_retry
    def recurs_get(url, headers):
        """
        Recursive function to handle pagination
        """
        try:
            response = requests.get(url, headers=headers, timeout=30)
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise Exception(
                f"Error {e.response.status_code}, could not get sharepoint list data: {e}"
            )
        except requests.exceptions.RequestException as e:
            raise Exception(f"Error, could not get sharepoint list data: {e}")

        data = response.json()

        # Check for the next page
        if "@odata.nextLink" in data:
            return data["value"] + recurs_get(data["@odata.nextLink"], headers)
        else:
            return data["value"]

    result = recurs_get(
        os.environ["GRAPH_BASE_URL"] + site_id + "/lists",
        headers={"Authorization": "Bearer " + os.environ["GRAPH_BEARER_TOKEN"]},
    )

    return result


@Decorators._refresh_graph_token
def get_sp_list_items(site_id: str, list_id: str, filter_query: str = None) -> dict:
    """
    Gets field data from a sharepoint list

    Note: If you're using the filter_query expression, whichever field you
    want to filter on needs to be indexed or you'll get an error.
    To index a column, just add it in the sharepoint list settings.

    :param site_id: The site id to get lists from
    :param list_id: The list id to get items from
    :param filter_query: An optional OData filter query
    """

    if "GRAPH_BASE_URL" not in os.environ:
        raise Exception("Error, could not find GRAPH_BASE_URL in env")

    @_basic_retry
    def recurs_get(url, headers):
        """
        Recursive function to handle pagination
        """
        try:
            response = requests.get(url, headers=headers, timeout=30)
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise Exception(
                f"Error {e.response.status_code}, could not get sharepoint list data: {e}"
            )
        except requests.exceptions.RequestException as e:
            raise Exception(f"Error, could not get sharepoint list data: {e}")

        data = response.json()

        # Check for the next page
        if "@odata.nextLink" in data:
            return data["value"] + recurs_get(data["@odata.nextLink"], headers)
        else:
            return data["value"]

    url = (
        os.environ["GRAPH_BASE_URL"]
        + site_id
        + "/lists/"
        + list_id
        + "/items?expand=fields"
    )

    if filter_query:
        url += "&$filter=" + filter_query

    result = recurs_get(
        url,
        headers={
            "Authorization": "Bearer " + os.environ["GRAPH_BEARER_TOKEN"],
            "Prefer": "HonorNonIndexedQueriesWarningMayFailRandomly",
        },
    )

    return result


@_basic_retry
@Decorators._refresh_graph_token
def get_sp_list_item(site_id: str, list_id: str, item_id: str) -> dict:
    """
    Gets field data from a specific sharepoint list item

    :param site_id: The site id to get lists from
    :param list_id: The list id to get items from
    :param item_id: The id of the list item to get field data from
    """
    if "GRAPH_BASE_URL" not in os.environ:
        raise Exception("Error, could not find GRAPH_BASE_URL in env")

    url = (
        os.environ["GRAPH_BASE_URL"]
        + site_id
        + "/lists/"
        + list_id
        + "/items/"
        + item_id
    )

    try:
        response = requests.get(
            url,
            headers={
                "Authorization": "Bearer " + os.environ["GRAPH_BEARER_TOKEN"],
                "Prefer": "HonorNonIndexedQueriesWarningMayFailRandomly",
            },
            timeout=30,
        )
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise Exception(
            f"Error {e.response.status_code}, could not get sharepoint list data: {e}"
        )
    except requests.exceptions.RequestException as e:
        raise Exception(f"Error, could not get sharepoint list data: {e}")

    return response.json()


@Decorators._refresh_graph_token
def create_sp_item(site_id: str, list_id: str, field_data: dict) -> dict:
    """
    Create a new item in SharePoint

    :param site_id: The site id to create the item in
    :param list_id: The list id to create the item in
    :param field_data: A dictionary of field data to create the item with, recommended
    to pull a list of fields from the list first to get the correct field names
    """
    try:
        response = requests.post(
            os.environ["GRAPH_BASE_URL"] + site_id + "/lists/" + list_id + "/items",
            headers={"Authorization": "Bearer " + os.environ["GRAPH_BEARER_TOKEN"]},
            json={"fields": field_data},
            timeout=30,
        )
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise Exception(
            f"Error {e.response.status_code}, could not create item in sharepoint: {e}"
        )
    except requests.exceptions.RequestException as e:
        raise Exception(f"Error, could not create item in sharepoint: {e}")

    return response.json()


@_basic_retry
@Decorators._refresh_graph_token
def delete_sp_item(site_id: str, list_id: str, item_id: str):
    """
    Delete an item in SharePoint

    :param site_id: The site id to delete the item from
    :param list_id: The list id to delete the item from
    :param item_id: The id of the list item to delete
    """
    try:
        response = requests.delete(
            os.environ["GRAPH_BASE_URL"]
            + site_id
            + "/lists/"
            + list_id
            + "/items/"
            + item_id,
            headers={"Authorization": "Bearer " + os.environ["GRAPH_BEARER_TOKEN"]},
            timeout=30,
        )
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise Exception(
            f"Error {e.response.status_code}, could not delete item in sharepoint: {e}"
        )
    except requests.exceptions.RequestException as e:
        raise Exception(f"Error, could not delete item in sharepoint: {e}")


@Decorators._refresh_graph_token
def update_sp_item(
    site_id: str, list_id: str, item_id: str, field_data: Dict[str, Any]
):
    """
    Update an item in SharePoint

    :param site_id: The site id to update the item in
    :param list_id: The list id to update the item in
    :param item_id: The id of the list item to update
    :param field_data: A dictionary of field data to update the item with, only
    include fields you're updating. Recommended to pull a list of fields from the list first to get the correct field names
    """
    try:
        response = requests.patch(
            os.environ["GRAPH_BASE_URL"]
            + site_id
            + "/lists/"
            + list_id
            + "/items/"
            + item_id
            + "/fields",
            headers={"Authorization": "Bearer " + os.environ["GRAPH_BEARER_TOKEN"]},
            json=field_data,
            timeout=30,
        )
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise Exception(
            f"Error {e.response.status_code}, could not update item in sharepoint: {e}"
        )
    except requests.exceptions.RequestException as e:
        raise Exception(f"Error, could not update item in sharepoint: {e}")


@Decorators._refresh_sp_token
def get_list_attachments(
    site_url: str, list_name: str, item_id: int, download: bool = False
) -> list[dict]:
    """
    Gets attachments for a sharepoint list item. Returns as a list of
    dicts (if the given list item does have attachments) if download is False.
    In other words, just downloading info about the attachments.

    Note: Uses the Sharepoint REST API, and not the Graph API.

    :param site_url: The site url to get list attachments from
    :param item_id: The id of the list item to get attachments from
    :param download: If True, download the attachments to the local filesystem
    """
    # Ensure the required environment variable is set
    if "SP_BEARER_TOKEN" not in os.environ:
        raise Exception("Error, could not find SP_BEARER_TOKEN in env")

    # Construct the URL for the ensure user endpoint
    url = f"{site_url}/_api/lists/getByTitle('{list_name}')/items({item_id})?$select=AttachmentFiles,Title&$expand=AttachmentFiles"

    try:
        response = requests.get(
            url,
            headers={
                "Authorization": "Bearer " + os.environ["SP_BEARER_TOKEN"],
                "Accept": "application/json;odata=verbose;charset=utf-8",
                "Content-Type": "application/json;odata=verbose;charset=utf-8",
            },
            timeout=30,
        )
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise Exception(
            f"Error {e.response.status_code}, could not get list attachments: {e}"
        )
    except requests.exceptions.RequestException as e:
        raise Exception(f"Error, could not get list attachments: {e}")

    # Get the attachment data
    data = response.json().get("d", {})
    attachments = data.get("AttachmentFiles", {}).get("results", [])

    if not download:
        return [
            {"name": str(x.get("FileName")), "url": str(x.get("ServerRelativeUrl"))}
            for x in attachments
        ]

    @_basic_retry
    def download_attachment(attachment):
        """
        Helper function to download an attachment
        """
        relative_url = attachment.get("ServerRelativeUrl")
        try:
            attachment_response = requests.get(
                f"{site_url}/_api/Web/GetFileByServerRelativeUrl('{relative_url}')/$value",
                headers={
                    "Authorization": "Bearer " + os.environ["SP_BEARER_TOKEN"],
                    "Accept": "application/json;odata=verbose;charset=utf-8",
                    "Content-Type": "application/json;odata=verbose;charset=utf-8",
                },
                timeout=30,
            )
            attachment_response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise Exception(
                f"Error {e.response.status_code}, could not download attachment: {e}"
            )
        except requests.exceptions.RequestException as e:
            raise Exception(f"Error, could not download attachment: {e}")

        return {
            "name": attachment.get("FileName"),
            "url": attachment.get("ServerRelativeUrl"),
            "data": attachment_response.content,
        }

    downloaded_files = []
    for attachment in attachments:
        downloaded_files.append(download_attachment(attachment))

    return downloaded_files


@Decorators._refresh_sp_token
def get_file(file_url: str) -> dict:
    """
    Downloads a file from a SharePoint site, likely stored in a document library.

    :param file_url: The direct URL to the file in the SharePoint document library
    :return: A dictionary containing the file name, URL, and file content
    """
    if "SP_BEARER_TOKEN" not in os.environ:
        raise Exception("Error, could not find SP_BEARER_TOKEN in env")

    headers = {
        "Authorization": "Bearer " + os.environ["SP_BEARER_TOKEN"],
        "Accept": "application/json;odata=verbose;charset=utf-8",
        "Content-Type": "application/json;odata=verbose;charset=utf-8",
    }

    # Parse the file URL to get the site URL and relative URL
    parsed_url = urlparse(file_url)
    path_parts = parsed_url.path.split("/")
    site_path = "/".join(path_parts[:3])  # This will include the site path
    relative_url = "/".join(path_parts[3:])  # This will include the rest of the path

    site_url = f"{parsed_url.scheme}://{parsed_url.netloc}{site_path}"

    print(f"Site URL: {site_url}")
    print(f"Relative URL: {relative_url}")

    try:
        response = requests.get(
            f"{site_url}/_api/Web/GetFileByUrl(@url)?@url='{file_url}'",
            headers=headers,
            timeout=30,
        )
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise Exception(f"Error {e.response.status_code}, could not download file: {e}")
    except requests.exceptions.RequestException as e:
        raise Exception(f"Error, could not download file: {e}")

    file_name = relative_url.split("/")[-1]

    return {"name": file_name, "url": file_url, "data": response.content}



"""
This module contains functions for interacting with users in MS Graph, both
actual users and also the site-specific users that are stored in a hidden
sharepoint list.
"""

import os
from typing import Optional

import requests

from grafap._auth import Decorators
from grafap._helpers import _basic_retry, _fetch_page


@Decorators._refresh_graph_token
def get_ad_users(select: str = None, filter: str = None, expand: str = None) -> dict:
    """
    Gets AD users in a given tenant

    :param select: OData $select query option
    :param filter: OData $filter query option
    :param expand: OData $expand query option
    """
    if "GRAPH_BASE_URL" not in os.environ:
        raise Exception("Error, could not find GRAPH_BASE_URL in env")

    @_basic_retry
    def recurs_get(url, headers):
        """
        Recursive function to handle pagination
        """
        try:
            response = requests.get(url, headers=headers, timeout=30)
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise Exception(
                f"Error {e.response.status_code}, could not get user data: {e}"
            )
        except requests.exceptions.RequestException as e:
            raise Exception(f"Error, could not get user data: {e}")

        data = response.json()

        # Check for the next page
        if "@odata.nextLink" in data:
            return data["value"] + recurs_get(data["@odata.nextLink"], headers)
        else:
            return data["value"]

    # Construct the query string
    query_params = []
    if select:
        query_params.append(f"$select={select}")
    if filter:
        query_params.append(f"$filter={filter}")
    if expand:
        query_params.append(f"$expand={expand}")

    query_string = "&".join(query_params)
    base_url = "https://graph.microsoft.com/v1.0/users"
    url = f"{base_url}?{query_string}" if query_string else base_url

    result = recurs_get(
        url,
        headers={"Authorization": "Bearer " + os.environ["GRAPH_BEARER_TOKEN"]},
    )

    return result


@Decorators._refresh_graph_token
def get_all_sp_users_info(site_id: str) -> dict:
    """
    Query the hidden sharepoint list that contains user information
    Can use "root" as the site_id for the root site, otherwise use the site id
    You would want to use whichever site ID is associated with the list you are querying

    :param site_id: The site id to get user information from
    """
    if "GRAPH_BASE_URL" not in os.environ:
        raise Exception("Error, could not find GRAPH_BASE_URL in env")

    @_basic_retry
    def recurs_get(url, headers, params=None):
        """
        Recursive function to handle pagination
        """
        try:
            response = requests.get(
                url,
                headers=headers,
                timeout=30,
                params=params,
            )
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise Exception(
                f"Error {e.response.status_code}, could not get sharepoint list data: {e}"
            )
        except requests.exceptions.RequestException as e:
            raise Exception(f"Error, could not get sharepoint list data: {e}")

        data = response.json()

        # Check for the next page
        if "@odata.nextLink" in data:
            return data["value"] + recurs_get(data["@odata.nextLink"], headers)
        else:
            return data["value"]

    url = (
        os.environ["GRAPH_BASE_URL"] + site_id + "/lists('User Information List')/items"
    )

    result = recurs_get(
        url,
        headers={"Authorization": "Bearer " + os.environ["GRAPH_BEARER_TOKEN"]},
        params={"expand": "fields(select=Id,Email)"},
    )

    return result


@_basic_retry
@Decorators._refresh_graph_token
def get_sp_user_info(
    site_id: str, user_id: Optional[str], email: Optional[str]
) -> dict:
    """
    Get a specific user from the hidden sharepoint list that contains user information

    :param site_id: The site id to get user information from
    :param user_id: The user id to get information for
    :param email: The email to get information for
    """
    if "GRAPH_BASE_URL" not in os.environ:
        raise Exception("Error, could not find GRAPH_BASE_URL in env")

    url = (
        os.environ["GRAPH_BASE_URL"] + site_id + "/lists('User Information List')/items"
    )

    if user_id:
        url += "/" + user_id
    elif email:
        url += "?$filter=fields/UserName eq '" + email + "'"

    try:
        response = requests.get(
            url,
            headers={
                "Authorization": "Bearer " + os.environ["GRAPH_BEARER_TOKEN"],
                "Prefer": "HonorNonIndexedQueriesWarningMayFailRandomly",
            },
            timeout=30,
        )
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise Exception(
            f"Error {e.response.status_code}, could not get sharepoint list data: {e}"
        )
    except requests.exceptions.RequestException as e:
        raise Exception(f"Error, could not get sharepoint list data: {e}")

    if "value" in response.json():
        if len(response.json()["value"]) == 0:
            raise Exception("Error, could not find user in sharepoint list")
        else:
            return response.json()["value"][0]
    return response.json()


# Doesn't seem to be needed, commenting out for now
# @Decorators.refresh_sp_token
# def get_site_user_by_id(site_url: str, user_id: str) -> dict:
#     """
#     Gets a sharepoint site user by the lookup id
#     """
#     headers = {
#         "Authorization": "Bearer " + os.environ["SP_BEARER_TOKEN"],
#         "Accept": "application/json;odata=verbose",
#     }

#     url = f"{site_url}/_api/web/siteusers/getbyid({user_id})"

#     response = requests.get(url, headers=headers, timeout=30)

#     if response.status_code != 200:
#         raise Exception("Error, could not get site user data: " + str(response.content))


@Decorators._refresh_sp_token
def ensure_sp_user(site_url: str, logon_name: str) -> dict:
    """
    Users sharepoint REST API, not MS Graph API. Endpoint is only available
    in the Sharepoint one. Ensure a user exists in given website. This is used
    so that the user can be used in sharepoint lists in that site. If the user has
    never interacted with the site or been picked in a People field, they are not
    available in the Graph API to pick from.

    :param site_url: The site url
    :param logon_name: The user's logon name, i.e. email address
    """
    # Ensure the required environment variable is set
    if "SP_BEARER_TOKEN" not in os.environ:
        raise Exception("Error, could not find SP_BEARER_TOKEN in env")

    # Construct the URL for the ensure user endpoint
    url = f"{site_url}/_api/web/ensureuser"

    # Make the POST request
    response = requests.post(
        url,
        headers={
            "Authorization": "Bearer " + os.environ["SP_BEARER_TOKEN"],
            "Accept": "application/json;odata=verbose;charset=utf-8",
            "Content-Type": "application/json;odata=verbose;charset=utf-8",
        },
        json={"logonName": logon_name},
        timeout=30,
    )

    # Check for errors in the response
    if response.status_code != 200:
        raise Exception(
            f"Error {response.status_code}, could not ensure user: {response.content}"
        )

    # Return the JSON response
    return response.json()



# # @Decorators.refresh_graph_token_delegated
# def get_user_notebooks():
#     """
#     Gets all notebooks for the signed-in user
#     """
#     if "GRAPH_BASE_URL" not in os.environ:
#         raise Exception("Error, could not find GRAPH_BASE_URL in env")

#     response = requests.get(
#         os.environ["GRAPH_BASE_URL"] + "me/onenote/notebooks",
#         headers={
#             "Authorization": "Bearer " + os.environ["USER_BEARER_TOKEN"],
#             "Accept": "application/json",
#         },
#         timeout=30,
#     )

#     if response.status_code != 200:
#         raise Exception("Error, could not get user notebooks: " + str(response.content))

#     return response.json()


# # @Decorators.refresh_graph_token_delegated
# def get_notebook_sections(notebook_id: str):
#     """
#     Gets all sections in a notebook
#     """
#     if "GRAPH_BASE_URL" not in os.environ:
#         raise Exception("Error, could not find GRAPH_BASE_URL in env")

#     response = requests.get(
#         os.environ["GRAPH_BASE_URL"]
#         + "me/onenote/notebooks/"
#         + notebook_id
#         + "/sections",
#         headers={
#             "Authorization": "Bearer " + os.environ["USER_BEARER_TOKEN"],
#             "Accept": "application/json",
#         },
#         timeout=30,
#     )

#     if response.status_code != 200:
#         raise Exception(
#             "Error, could not get notebook sections: " + str(response.content)
#         )

#     return response.json()


# # @Decorators.refresh_graph_token_delegated
# def get_section_pages(notebook_id: str, section_id: str):
#     """
#     Gets all pages in a section
#     """
#     if "GRAPH_BASE_URL" not in os.environ:
#         raise Exception("Error, could not find GRAPH_BASE_URL in env")

#     response = requests.get(
#         os.environ["GRAPH_BASE_URL"]
#         + "me/onenote/notebooks/"
#         + notebook_id
#         + "/sections/"
#         + section_id
#         + "/pages",
#         headers={
#             "Authorization": "Bearer " + os.environ["USER_BEARER_TOKEN"],
#             "Accept": "application/json",
#         },
#         timeout=30,
#     )

#     if response.status_code != 200:
#         raise Exception("Error, could not get section pages: " + str(response.content))

#     return response.json()
